@@ -1,21 +1,27 @@
+#[cfg(unix)]
+use command_group::UnixChildExt;
+use command_group::{CommandGroup, GroupChild};
+use log::{debug, info, warn};
+
+use std::{
+    collections::HashMap,
+    fs::canonicalize,
+    process::Child,
+    sync::{
+        mpsc::{channel, Receiver},
+        Arc, Mutex,
+    },
+    time::Duration,
+};
+
 use crate::config::Config;
 use crate::error::{Error, Result};
 use crate::gitignore;
 use crate::ignore;
 use crate::notification_filter::NotificationFilter;
 use crate::pathop::PathOp;
-use crate::process::{self, Process};
 use crate::signal::{self, Signal};
 use crate::watcher::{Event, Watcher};
-use std::{
-    collections::HashMap,
-    fs::canonicalize,
-    sync::{
-        mpsc::{channel, Receiver},
-        Arc, RwLock,
-    },
-    time::Duration,
-};
 
 /// Behaviour to use when handling updates while the command is running.
 #[derive(Clone, Copy, Debug)]
@@ -138,15 +144,79 @@ where
     Ok(())
 }
 
+#[derive(Debug)]
+pub enum ChildProcess {
+    None,
+    Grouped(GroupChild),
+    Ungrouped(Child),
+}
+
+impl Default for ChildProcess {
+    fn default() -> Self {
+        ChildProcess::None
+    }
+}
+
+impl ChildProcess {
+    #[cfg(unix)]
+    fn signal(&mut self, sig: Signal) -> Result<()> {
+        match self {
+            Self::None => Ok(()),
+            Self::Grouped(c) => {
+                debug!("Sending signal {} to process group id={}", sig, c.id());
+                c.signal(sig)
+            }
+            Self::Ungrouped(c) => {
+                debug!("Sending signal {} to process id={}", sig, c.id());
+                c.signal(sig)
+            }
+        }
+        .map_err(|e| e.into())
+    }
+
+    fn kill(&mut self) -> Result<()> {
+        match self {
+            Self::None => Ok(()),
+            Self::Grouped(c) => {
+                debug!("Killing process group id={}", c.id());
+                c.kill()
+            }
+            Self::Ungrouped(c) => {
+                debug!("Killing process id={}", c.id());
+                c.kill()
+            }
+        }
+        .map_err(|e| e.into())
+    }
+
+    fn is_running(&mut self) -> Result<bool> {
+        match self {
+            Self::None => Ok(false),
+            Self::Grouped(c) => c.try_wait().map(|w| w.is_none()),
+            Self::Ungrouped(c) => c.try_wait().map(|w| w.is_none()),
+        }
+        .map_err(|e| e.into())
+    }
+
+    fn wait(&mut self) -> Result<()> {
+        match self {
+            Self::None => Ok(()),
+            Self::Grouped(c) => c.wait().map(drop),
+            Self::Ungrouped(c) => c.wait().map(drop),
+        }
+        .map_err(|e| e.into())
+    }
+}
+
 pub struct ExecHandler {
     args: Config,
     signal: Option<Signal>,
-    child_process: Arc<RwLock<Option<Process>>>,
+    child_process: Arc<Mutex<ChildProcess>>,
 }
 
 impl ExecHandler {
     pub fn new(args: Config) -> Result<Self> {
-        let child_process: Arc<RwLock<Option<Process>>> = Arc::new(RwLock::new(None));
+        let child_process: Arc<Mutex<ChildProcess>> = Arc::default();
         let weak_child = Arc::downgrade(&child_process);
 
         // Convert signal string to the corresponding integer
@@ -154,11 +224,21 @@ impl ExecHandler {
 
         signal::install_handler(move |sig: Signal| {
             if let Some(lock) = weak_child.upgrade() {
-                let strong = lock.read().expect("poisoned lock in install_handler");
-                if let Some(ref child) = *strong {
-                    match sig {
-                        Signal::SIGCHLD => child.reap(), // SIGCHLD is special, initiate reap()
-                        _ => child.signal(sig),
+                let mut child = lock.lock().expect("poisoned lock in install_handler");
+                match sig {
+                    Signal::SIGCHLD => {
+                        child.is_running().ok();
+                    }
+                    _ => {
+                        #[cfg(unix)]
+                        child.signal(sig).unwrap_or_else(|err| {
+                            warn!("Could not pass on signal to command: {}", err)
+                        });
+
+                        #[cfg(not(unix))]
+                        child.kill().unwrap_or_else(|err| {
+                            warn!("Could not pass on termination to command: {}", err)
+                        });
                     }
                 }
             }
@@ -176,25 +256,34 @@ impl ExecHandler {
             clearscreen::clear()?;
         }
 
+        let mut child = self.child_process.lock()?;
+        child.kill().ok();
+
+        let mut command = self.args.shell.to_command(&self.args.cmd);
+        debug!("Assembled command: {:?}", command);
+
+        if !self.args.no_environment {
+            for (name, val) in crate::paths::collect_path_env_vars(ops) {
+                debug!("Command environment: {}={:?}", name, val);
+                command.env(name, val);
+            }
+        }
+
         debug!("Launching command");
-        let mut guard = self.child_process.write()?;
-        *guard = Some(process::spawn(
-            &self.args.cmd,
-            ops,
-            self.args.shell.clone(),
-            !self.args.no_environment,
-        )?);
+        *child = if self.args.use_process_group {
+            ChildProcess::Grouped(command.group_spawn()?)
+        } else {
+            ChildProcess::Ungrouped(command.spawn()?)
+        };
 
         Ok(())
     }
 
-    pub fn has_running_process(&self) -> bool {
-        let guard = self
-            .child_process
-            .read()
-            .expect("poisoned lock in has_running_process");
-
-        (*guard).is_some()
+    pub fn has_running_process(&self) -> Result<bool> {
+        self.child_process
+            .lock()
+            .expect("poisoned lock in has_running_process")
+            .is_running()
     }
 }
 
@@ -213,29 +302,37 @@ impl Handler for ExecHandler {
         Ok(true)
     }
 
-    // Only returns Err() on lock poisoning.
     fn on_update(&self, ops: &[PathOp]) -> Result<bool> {
-        let signal = self.signal.unwrap_or(Signal::SIGTERM);
+        log::debug!("ON UPDATE: called");
 
-        match (self.has_running_process(), self.args.on_busy_update) {
+        let signal = self.signal.unwrap_or(Signal::SIGTERM);
+        let has_running_processes = self.has_running_process()?;
+
+        log::debug!(
+            "ON UPDATE: has_running_processes: {} --- on_busy_update: {:?}",
+            has_running_processes,
+            self.args.on_busy_update
+        );
+
+        match (has_running_processes, self.args.on_busy_update) {
             // If nothing is running, start the command
             (false, _) => {
                 self.spawn(ops)?;
             }
 
             // Just send a signal to the command, do nothing more
-            (true, OnBusyUpdate::Signal) => signal_process(&self.child_process, signal),
+            (true, OnBusyUpdate::Signal) => signal_process(&self.child_process, signal)?,
 
             // Send a signal to the command, wait for it to exit, then run the command again
             (true, OnBusyUpdate::Restart) => {
-                signal_process(&self.child_process, signal);
-                wait_on_process(&self.child_process);
+                signal_process(&self.child_process, signal)?;
+                wait_on_process(&self.child_process)?;
                 self.spawn(ops)?;
             }
 
             // Wait for the command to end, then run it again
             (true, OnBusyUpdate::Queue) => {
-                wait_on_process(&self.child_process);
+                wait_on_process(&self.child_process)?;
                 self.spawn(ops)?;
             }
 
@@ -245,10 +342,10 @@ impl Handler for ExecHandler {
         // Handle once option for integration testing
         if self.args.once {
             if let Some(signal) = self.signal {
-                signal_process(&self.child_process, signal);
+                signal_process(&self.child_process, signal)?;
             }
 
-            wait_on_process(&self.child_process);
+            wait_on_process(&self.child_process)?;
 
             return Ok(false);
         }
@@ -316,20 +413,25 @@ fn wait_fs(
     paths
 }
 
-fn signal_process(process: &RwLock<Option<Process>>, signal: Signal) {
-    let guard = process.read().expect("poisoned lock in signal_process");
+fn signal_process(process: &Mutex<ChildProcess>, signal: Signal) -> Result<()> {
+    let mut child = process.lock().expect("poisoned lock in signal_process");
 
-    if let Some(ref child) = *guard {
-        debug!("Signaling process with {}", signal);
-        child.signal(signal);
+    #[cfg(unix)]
+    child.signal(signal)?;
+
+    #[cfg(not(unix))]
+    if matches!(signal, Signal::SIGTERM | Signal::SIGKILL) {
+        child.kill()?;
+    } else {
+        debug!("Ignoring signal to send to process");
     }
+
+    Ok(())
 }
 
-fn wait_on_process(process: &RwLock<Option<Process>>) {
-    let guard = process.read().expect("poisoned lock in wait_on_process");
-
-    if let Some(ref child) = *guard {
-        debug!("Waiting for process to exit...");
-        child.wait();
-    }
+fn wait_on_process(process: &Mutex<ChildProcess>) -> Result<()> {
+    process
+        .lock()
+        .expect("poisoned lock in wait_on_process")
+        .wait()
 }
